(*************** test framework ********************)

(* Example routine to modify the text to uppercase *)
let upc editor ()= 
	let xs= Wid.get_contents editor in
	let us= String.uppercase xs in
	Wid.set_contents us editor;;

(* Example routine to modify the text to lowercase *)
let toLower editor ()= 
let xs= Wid.get_contents editor in
	let us= String.lowercase xs in
	Wid.set_contents us editor;;

let nxt_mask editor ()= 
let us= Wid.the_mask ["Sells:########\013Profit ########"; "Cost:########     "]  in
		Wid.set_contents us editor;;	

(* Example routine to save the text to a file *)
let counter= ref 0;;
let savit editor ()=
	let s= Wid.get_contents editor and
		chan= open_out (editor^".b"^(string_of_int !counter)) in
	incr counter;
	output_string chan s;
	close_out chan;;

(* Example routine to calculete the mean of a list of numbers *)
let avg s= 
	let rec loop nums acc c= match nums with
		  [] -> acc /. c
		| x::xs -> loop xs (acc +. x) (c +. 1.0) in
	loop s 0.0 0.0;;

(* returns just the numbers in float format *)
let rec rdThem s=
  try
    match Stream.next s with
      | Genlex.Int x -> (float_of_int x)::(rdThem s)
      | Genlex.Float x -> x::(rdThem s)
      | _  -> rdThem s
  with
     _ -> [];;

(* get a list of float using ocamlex *)
let getFloatList s=
  let lex= Genlex.make_lexer [] in
    let strm= lex (Stream.of_string s) in
         rdThem strm;;

(* process the text and return the text result *)
let find_avg s=
	let xs= getFloatList s in
		match xs with
			  [] -> "Empty list"
			| nums -> "The average is "^(string_of_float (avg nums));;

(* just a shell to process the text with the mask *)
let processa msk s= find_avg(Wid.get_fields msk s);;

let main () = 
	Wid.open_gr "eMacs like Editor - Gui Example";
	let msk= Wid.the_mask ["Product:########\013Price ########"; "Quantity:########     "] in
	let ed = Wid.mk_edt ~txt:(Wid.nlines 20) ~fn:(fun x -> (String.uppercase x)) "Editor"  and
		 ed2 = Wid.mk_entry ~txt:msk ~fn:(processa msk) "Form"  in
	let ext= Wid.exit_button() and
		 uppr= Wid.button 0 (upc "Editor") "Uppercase" and
		 lc= Wid.button 1 (toLower "Editor") "Lowercase"  and
		 sv= Wid.next_button_row 0 (savit "Editor")  "Save" and 
		 chgmsk= Wid.next_button_row 1 (nxt_mask "Form") "Process"
	in
	while true do
		Wid.refresh ();
		ed.Wid.process_event();
		ed2.Wid.process_event ();
		uppr.Wid.process_event();
		lc.Wid.process_event();
		chgmsk.Wid.process_event();
		sv.Wid.process_event();
		ext.Wid.process_event()	
	done;
	Wid.close_gr();;   
 
main ();;