(* Type declaration section *)

(* Rectangle type definition (x,y) - position on the screen (w,h)=(width,heigth) of the rectangle *)
type box_config = { x:int; y:int; w:int; h:int };;

(* eMacs like editor object type definition *)
type txt ={
(* Text of the Editor *)
				mutable tx: string; 
(* Position of the line and cursor *)
				mutable cpos:(int*int); 
(* Position of the cursor in the current line *)
				mutable stx:int; 
(* Position of the first position to display in the window *)
				mutable sty:int;
(* Used to mark a positon in the text *)
				mutable mark:int;
				mutable mask: string };;

(* Image or colour  on the button *)
type widg= IMG of Graphics.image | CLR of Graphics.color;;

(* Process definition *)
type widget =
{		process_event : (unit -> unit);
		st: (txt -> widg -> bool -> unit);
		draw_self : (unit -> unit) }

(* Structure to storage all local variables needed in the text editor *)
type edit_state = {  
(* Backup stx used some times *)
				mutable bkpx:int; 
(* name of the Editor *)
				mutable nm:string; 
(* Used to activate or deactivate the erase of the caractere on backspace key *)
				mutable bkerase:bool;
(* buffer to safe the last test *)
				mutable ctx_save:string;
(* buffer to safe the last mask *)
				mutable cmk_save:string;
(*  *)
				mutable sty_stk: int list; 
(*  *)
				mutable insertions:bool;
(* All main variables of the Editor *)
				mutable etx:txt;
(* Used to storage a virtual key  *)
				mutable virtual_char:int}

(* structure to store all global variables for all objects *)
type glb= {  
(* Define the number of lines showed in the window *)
				 mutable lines_per_page:int; 
(* Define the number of columns showed in the window *)
				 mutable chars_per_line:int;
(* Mouse position when an event occors *)
				 mutable mouse_position:(int*int);
(* Yank definition its the list buffer that storage parts of the text on memory *)
				 mutable yank:(string list*string list);
(* list of componets to rebuild its graphics elements *)
				 mutable refresh_list:widget list;
(* The last position find for any search done in editors *)
				 mutable search_pos:int;
(* Control a virtual keypress *)
				 mutable virtual_keys:int list;
				 mutable mrk:int};;


(* Main variables definitions *)

(* Global variables initialization *)
let glb= {	lines_per_page= 5; 
		chars_per_line= 10; 
		mouse_position= (0,0); 
		yank=([], []);
		refresh_list= []; 
		search_pos= 0;
		virtual_keys=[];
		mrk=(-1)};;

(* Text editor controls *)
(* The number of Editors created *)
let 	top= ref 0 and
(* setup to control only 5 editors *)
(* Stack to storage the name of the editor *)
	pilha= Array.make 5 "" and
(* The structure to storage the contents of the Editor *)
	oldStuff= Array.make 5 "" and
(* The structure to storage the contents of the mask for all Editors *)
	mskStuff= Array.make 5 "" and
(* create positions in the memory for 5 editors *)
	eds= Array.make 5 { process_event= (fun () -> ()); 
				  st= (fun x y z -> ()); 
				  draw_self=(fun () -> ())} and
(* border size on the main window *)
		winHeight=5;;

(* Define grey levels *)
let set_gray x = (Graphics.rgb x x x);;
(* Define several tones of grey and white used in Editors and buttons *)
let	gray1 	= set_gray 100 and 
	gray2	= set_gray 170 and 
	gray3	= set_gray 240 and
	background_mask	= set_gray 215 and
	background_edt	= set_gray 200;;

let last_sz= ref (0, 0);;

let set_siz ()= last_sz := (Graphics.size_x(), Graphics.size_y());;


(* Routines *)

(* get the top of the circular list yank-buffer *)
let getop ()= 
 try
  match glb.yank with
      (ys, []) -> glb.yank <- ([], List.rev ys); List.hd ys
    | (ys, (x::xs)) -> x
 with
    any -> "";;
(* circle the list yank-buffer *)
let circula ()= 
 try
  match glb.yank with
       ([], [x]) -> ()
    |  (ys, [x]) -> glb.yank <- ([x], List.rev ys)
    |  (ys, (x::xs)) -> glb.yank <- ((x::ys), xs)
    
    | (ys, []) -> glb.yank <- ([], List.rev ys)
 with
   any -> ();;
(* push a string on the list yank-buffer *)
let push_circular x = 
    let (xs, ys)= glb.yank in
        glb.yank <- (xs, x::ys);;

(* constructor of the rectangle *)
let mkrec ax ay wx wy= {x= ax; y=ay; w=wx; h=wy};;
(* constructor of the eMac like editor *)

let mask_fun s i=
  let last_line= String.length s - 1 in
          if i <0 || i > last_line then false
          else ( s.[i]= '#' );;
          
let createText xs = {cpos = (0, 0);  tx = xs; stx=0; sty=0; mark=0; mask= ""};;

         
let createMaskText xs msk = {cpos = (0, 0);  tx =xs ; stx=0; sty=0; mark=0; 
                              mask=msk};;

(* integer Stack controls *)
let clear_stack stack = stack := [] and
     empty_stack stack = match stack.contents with
	[] -> true
	| _ -> false
and push stack data =
	stack:=data::stack.contents and
    pop stack = match stack.contents with
	[] -> 0
	| (head::tail) -> stack:=tail; head ;;

(* invert the index of a string *)
let gch s i =
  let sz= String.length s - 1 in
        s.[sz - i];;

(* search string using Boyer/More algoritm *)
let search pat text index=
(* computation of the shift-table - phase 1 *)
	let m=String.length pat and 
	     n =String.length text in
	let rborder=Array.create (m+1) 0 and
	     d = Array.create (m+1) 1 and
	     last = Array.create 256 0 in
	for j = 0 to m-1 do
	    rborder.(j) <- 1;
	    d.(j) <- m;
	done;
	let j= ref 1 and i= ref (m-1) in
	while i.contents >= j.contents do 
	  while i.contents >= j.contents && pat.[m-j.contents]=pat.[i.contents-j.contents]  do
		j:=j.contents+1;
		rborder.(i.contents-j.contents+1) <- j.contents;
	  done;
	  if j.contents> 1 then (d.(m-j.contents+1) <- min (m- i.contents) d.(m-j.contents+1););
	  i:=i.contents-j.contents+rborder.(m-j.contents+1);
	  j:=max 1 rborder.(m-j.contents+1);
	done;
(* computation of the shift-table - phase 2 *)
	let t= ref rborder.(0) and l= ref 1 and s= ref 0 in
	while t.contents > 0 do
	  s:=m-t.contents+1;
		for j=l.contents to s.contents do
		  d.(j) <- min d.(j) s.contents;
		done;
	  t:= rborder.(s.contents);
	  l:=s.contents+1;
	done;
	for i=0 to m-1 do
	  last.(int_of_char pat.[i]) <- i+1;
	done;
(* Algorithm of Boyer/Moore *)
	i:=index; j:=m;
	while i.contents<=n-m+1 && j.contents>0 do
	  j:=m;
	  while j.contents >=1 && pat.[j.contents-1]=text.[i.contents+j.contents -2] do
		j:=j.contents-1;
	  done;
	  if j.contents> 0 then 
		  i:= i.contents+max d.(j.contents) (j.contents-last.(int_of_char text.[i.contents+j.contents-2])) ;
	done;
	if j.contents>0 then 0 else i.contents;;

(* search string backword using Boyer/More algoritm *)
let search_back pat text index=
(* computation of the shift-table - phase 1 *)
	let m=String.length pat and 
	     n =String.length text in
	let rborder=Array.create (m+1) 0 and
	     d = Array.create (m+1) 1 and
	     last = Array.create 256 0 in
	for j = 0 to m-1 do
	    rborder.(j) <- 1;
	    d.(j) <- m;
	done;
	let j= ref 1 and i= ref (m-1) in
	while i.contents >= j.contents do 
	  while i.contents >= j.contents && gch pat (m-j.contents)=gch pat (i.contents-j.contents)  do
		j:=j.contents+1;
		rborder.(i.contents-j.contents+1) <- j.contents;
	  done;
	  if j.contents> 1 then (d.(m-j.contents+1) <- min (m- i.contents) d.(m-j.contents+1););
	  i:=i.contents-j.contents+rborder.(m-j.contents+1);
	  j:=max 1 rborder.(m-j.contents+1);
	done;
(* computation of the shift-table - phase 2 *)
	let t= ref rborder.(0) and l= ref 1 and s= ref 0 in
	while t.contents > 0 do
	  s:=m-t.contents+1;
		for j=l.contents to s.contents do
		  d.(j) <- min d.(j) s.contents;
		done;
	  t:= rborder.(s.contents);
	  l:=s.contents+1;
	done;
	for i=0 to m-1 do
	  last.(int_of_char (gch pat i)) <- i+1;
	done;
(* Algorithm of Boyer/Moore *)
	i:=n-index; j:=m;
	while i.contents<=n-m+1 && j.contents>0 do
	  j:=m;
	  while j.contents >=1 && (gch pat (j.contents-1)) = (gch text (i.contents+j.contents -2)) do
		j:=j.contents-1;
	  done;
	  if j.contents> 0 then 
		  i:= i.contents+max d.(j.contents) (j.contents-last.(int_of_char (gch text (i.contents+j.contents-2)))) ;
	done;
	if j.contents>0 then 0 else n-i.contents;;

(* Graphics routines *)
(* Open a graphic window *)
let open_gr nm = 
   Graphics.open_graph "" ; 
   Graphics.set_window_title nm;;
(* Close a graphic window *)
let close_gr ()=
   Graphics.close_graph();;
(* draw a char (c) on the current graphics screen on the (x,y) position *)
let draw_char x y c= Graphics.moveto x y; Graphics.draw_char c;;
(* draw a string (s) on the current graphics screen on the (x,y) position *)
let draw_string x y s= Graphics.moveto x y; Graphics.draw_string s;;             

(* Resize the size of the char to extend the line *)
let 	line_space=0 and
	caractere_space=1;;
let resize_char x = (caractere_space+fst x, line_space+snd x);;

let clear_graph ()=
  let rec loop s= match s with
        [] -> ()
      | v::vs -> v.st {cpos= (0,0);  tx= ""; stx=0; sty=0; mark=0; mask= ""} 
             (CLR Graphics.red) true; v.draw_self(); 
	     loop vs
  in
      (  
         Graphics.auto_synchronize false; 
         Graphics.clear_graph();
         Graphics.set_color Graphics.black;
         loop glb.refresh_list;
         Graphics.auto_synchronize true);;   

let tag id rc=
		let (wt, ht)= (120, 25) in
		let (xc, yc)= (rc.x + id*wt, rc.y + rc.h) in 
		{x= xc; y= yc; w= wt-2; h= ht};;

(* draw a box button on bfc position and bfc size and bw borter size and colour t *)
let draw_box t bw bcf =
   let x1 = bcf.x and y1= bcf.y in
   let x2= x1+bcf.w and y2= y1+bcf.h in
   let ix1= x1 + bw and ix2 = x2 - bw and
       iy1= y1 + bw and iy2= y2 - bw in
   let border1 g=
            Graphics.set_color g;
            Graphics.fill_poly
                 [| (x1, y1); (ix1, iy1); (ix2, iy1); 
                    (ix2, iy2);  (x2, y2); (x2, y1) |] in
   let border2 g=
           Graphics.set_color g;
           Graphics.fill_poly
                [| (x1, y1); (ix1, iy1);
                   (ix1, iy2); (ix2, iy2);
                   (x2, y2); (x1, y2)|] in
   (match t= Graphics.red with
       true -> Graphics.set_color gray2;
              Graphics.fill_rect ix1 iy1 (ix2-ix1) (iy2-iy1);
              border1 gray1;
              border2 gray3
     | false -> Graphics.set_color gray2;
              Graphics.fill_rect ix1 iy1 (ix2-ix1) (iy2-iy1);
              border1 gray3;
              border2 gray1);
   Graphics.set_color Graphics.black;
   Graphics.draw_rect bcf.x bcf.y bcf.w bcf.h;;

(* Draw the text in a window *)
let drawText pos_x pos_y c=
(* pos_x-> x position of the window, pos_y -> y position of the window, c -> the editor *)
  let (char_w, char_h)= resize_char (Graphics.text_size "w")  in
  let rec loop i j pos=
(* i->line on the screen, j-> column on the screen, pos-> position in the string *)
  if (snd c.cpos)=pos then
    ( if j< glb.chars_per_line then (draw_char (pos_x+j*char_w-char_w/2) (pos_y-i*char_h) '|'; c.stx <- j; c.cpos <- ( i,snd c.cpos );)
      else if j= glb.chars_per_line then (draw_char (pos_x+j*char_w-char_w/2) (pos_y-i*char_h) '|'; c.stx <- j); c.cpos <- ( i,snd c.cpos ););
(* draw the cursor *)
   match pos with
	   position when (position >= String.length c.tx ) || 
	                 (i >= glb.lines_per_page) -> () 
(* if end of the string or line position greater than the window, than stop *)
	| position when (c.tx.[position]= '\013') -> draw_char (pos_x+j*char_w) (pos_y-i* char_h) c.tx.[pos]; loop (i+1) 0 (position+1)
(* if find a <ENTER> caractere than next line *)
	| position when (j > glb.chars_per_line) -> loop (i+1) 0 (position)
(* if column position greater than window than next line *)
	| _ -> ( if c.mark= pos then ( Graphics.set_color Graphics.red;
	                               draw_char (pos_x+j*char_w) (pos_y-i* char_h) c.tx.[pos];
	                               Graphics.set_color Graphics.black)
	         else (draw_char (pos_x+j*char_w) (pos_y-i* char_h) c.tx.[pos])); loop i (j+1) (pos+1)
(* or just print the caractere *)
 in 
    loop 0 0 c.sty;;

(* Draw the text in a window *)
let drawMask pos_x pos_y c=
(* pos_x-> x position of the window, pos_y -> y position of the window, c -> the editor *)
  let (char_w, char_h)= resize_char (Graphics.text_size "w")  in
  let rec loop i j pos=
(* i->line on the screen, j-> column on the screen, pos-> position in the string *)
let draw_cursor ()=  if (snd c.cpos)=pos then
    ( if j< glb.chars_per_line then (draw_char (pos_x+j*char_w-char_w/2) (pos_y-i*char_h) '|'; c.stx <- j; c.cpos <- ( i,snd c.cpos );)
      else if j= glb.chars_per_line then (draw_char (pos_x+j*char_w-char_w/2) (pos_y-i*char_h) '|'; c.stx <- j); c.cpos <- ( i,snd c.cpos ););
in
(* draw the cursor *)
   match pos with
	   position when (position >= String.length c.tx ) || 
	                 (i >= glb.lines_per_page) -> () 
(* if end of the string or line position greater than the window, than stop *)
	| position when (c.tx.[position]= '\013') -> draw_char (pos_x+j*char_w) (pos_y-i* char_h) c.tx.[pos];draw_cursor (); loop (i+1) 0 (position+1)
(* if find a <ENTER> caractere than next line *)
	| position when (j > glb.chars_per_line) ->draw_cursor (); loop (i+1) 0 (position)
(* if column position greater than window than next line *)
	| position when c.mask.[pos]='#' -> 
							  ( if c.tx.[pos]='#' then ( 	Graphics.set_color Graphics.white;
							    					Graphics.fill_rect (pos_x+j*char_w) (pos_y-i* char_h) char_w (char_h-1) ;
												Graphics.set_color Graphics.black;)
							    			    else ( 	Graphics.set_color Graphics.white;
							    			    		Graphics.draw_rect (pos_x+j*char_w) (pos_y-i* char_h) char_w (char_h-1) ;
							    			    		Graphics.set_color Graphics.black;
							    			    		draw_char (pos_x+j*char_w) (pos_y-i* char_h) c.tx.[pos];);
							   draw_cursor (); loop i (j+1) (pos+1); )
	| _ -> draw_char (pos_x+j*char_w) (pos_y-i* char_h) c.tx.[pos]; draw_cursor ();loop i (j+1) (pos+1)
	          
(* or just print the caractere *)
 in 
    loop 0 0 c.sty;;

(* Mouse routines *)
(* Mouse cursor in the box *)
let mouse_in_window (x, y) window =
   if ( (x >= window.x) && 
     (x < (window.x + window.w)) && (y >= window.y) && (y < (window.y + window.h))) 
   then true else false;;
(* Mouse cursor in the window ? *)
let click_in_window window =
  if Graphics.button_down() &&  mouse_in_window (Graphics.mouse_pos()) window 
  then true
  else false;;

(* button routines *)
(* Get the colour of the button *)
let get_color gcolor= match gcolor with
      (CLR c) -> c
    | any -> Graphics.red;;
(* Get the image of the button *)
let get_img image_defalt image = match image with
         (IMG i) -> i
       | any -> image_defalt;;

(* Create a string with n lines of blank text *)
let  nlines k= 
 let	(cw, ch)= resize_char (Graphics.text_size "w") and
		(size_x, size_y)= (Graphics.size_x(), Graphics.size_y()) in
  let	line_len = (size_x - 100) / cw in
		String.make ((k+1)*line_len) ' ' ;;

(* get the index of the Editor *)
let getId editor_name = 
  let rec getId_loop id editor_name= match id with
	  i when i > !top -> 0
	| i when pilha.(i)= editor_name -> i
	| i -> getId_loop (i+1) editor_name
  in getId_loop 0 editor_name;;

(* load a file into the id Editor *)
let load_file id nm=
   let acc= ref [] in
   let chan= open_in nm in
    ( try
        while true do
         acc := (input_line chan)::acc.contents
        done
      with 
       any -> close_in chan);
      (List.rev acc.contents);;

(* get the text content of the nm Editor *)
let get_contents editor_name= oldStuff.(getId editor_name);;

(* set the text content of the nm Editor with s *)
let set_contents s editor_name= 
 let id= getId editor_name in
   oldStuff.(id) <- s;
    eds.(id).st {cpos=(0,0); tx=oldStuff.(id); stx =0 ; sty=0; 
                      mark=0; mask="" } (CLR Graphics.red) true;
    eds.(id).draw_self();;

let set_mask s editor_name= 
 let id= getId editor_name in
   oldStuff.(id) <- s;
   mskStuff.(id) <- (String.copy s);
    eds.(id).st {cpos=(0,0); tx=oldStuff.(id); stx =0 ; sty=0; 
                      mark=0; mask= (String.copy s) } (CLR Graphics.red) true;
    eds.(id).draw_self();;
    

(* Change the name of the Editor nm to newnm *)
let change_name nm newnm=
  let id= getId nm in
     pilha.(id) <- newnm;
     let xs= if Sys.file_exists newnm then String.concat "" (load_file id newnm) 
                else String.make 5000 ' ' in
          (  (*num_of_lines.(id) := (String.length xs ) / 60;*)
             set_contents xs newnm);
      eds.(id).draw_self();;

(* Returns the stack of the id(th) element on the stack *)
let push label= pilha.(!top) <- label;
   let i= !top in
    ( (if !top < (Array.length pilha) - 1 then incr top);  i);;



(* Read just one line: Emacs data input *)

type line= {  line_input: (int -> int -> txt  -> unit); 
              state: (char list -> unit);
              draw_line: (unit -> unit)};;
type  t =
  { mutable before : char list;
    mutable after : char list};;
let xxx= ref "";;

let make_line_input  sz rc=
let (tw, th)= resize_char (Graphics.text_size "w")  in
let charWidth = ref 8 in
let stt = ref 0 and maxCharNum = ref sz in
let 
  c= {before = []; after = ['-']} and
  ch= ref ' ' and
  sc= ref [] and
  ascii = ref 0 and
  field_length= Graphics.size_x () - 50  in
let rec drawList pos j i s=
  match s with
     [] -> ()
   | x::xs when j <  !stt -> drawList pos (j+1) i xs
   | x::xs when i  <= !maxCharNum ->
        Graphics.moveto (i * !charWidth) pos; 
        Graphics.draw_char x; 
        drawList pos j (i+1) xs
   | x::xs -> Graphics.moveto (i * !charWidth) pos; 
              Graphics.set_color Graphics.red;
              Graphics.draw_char x;
              Graphics.set_color Graphics.black in
let draw_line () = 
	Graphics.set_color Graphics.white;
	Graphics.fill_rect rc.x rc.y rc.w rc.h;
	Graphics.set_color Graphics.black;
	Graphics.draw_rect rc.x rc.y rc.w rc.h;
	Graphics.moveto (25+tw) (rc.y+10);
	drawList (rc.y+th/2) 0 4 sc.contents  in
        
let backstep c =
  match c.before with
    [] -> ()
  | x :: l -> (
      c.after <- x::c.after;
      c.before <- l;
      if !stt > 0 then stt := !stt - 1
    ) in
  let step c =
     match c.after with
         [] -> ()
       | x :: l -> 
         (
           c.before <- x::c.before;
           c.after <- l;
           if List.length c.before > !maxCharNum then stt := !stt + 1
          )  in  
let delete c=
  match c.before with
      [] -> ()
    | x::xs -> c.before <- xs;
               if !stt > 0 
               then stt := !stt - 1 in
let insert c x =
   if (List.length c.before) >=  !maxCharNum then stt := !stt + 1; 
   c.before <- x::c.before in
let get_all c = (List.rev c.before) @ ('_':: c.after) in
let get_chars c = (List.rev c.before) @ c.after in

let get_string c=
  let
    i = ref 0 and
    s= ref (get_chars c) in
  let
      str= String.create (List.length  !s) in
    begin
      while !s <> [] do
        str.[!i] <- List.hd (!s);
        s := List.tl s.contents;
        incr i
      done;
      str
    end in
let savt nm s =
   let chan= open_out nm in
      output_string chan s; 
       close_out chan 
in let stpYes= ref true 
and enter= ref 0 
and erase_sc= ref false in
let read_string s_or_x id xs=
  maxCharNum := field_length / !charWidth - 1;
  ch := Graphics.read_key();
  c.before <- []; c.after <- [];
  stpYes := false;
  if (s_or_x= 1) then enter := 3 
  else (if s_or_x= 2 then enter := 4 else enter := 0);
  erase_sc := false;
  while enter.contents <> 5  do
         (*ch := Graphics.read_key(); *)
         if (int_of_char !ch) < 32 then
           ( ); 
         ( match (int_of_char !ch) with
              2  ->  backstep c
            | 14 -> step c  (* 6 is Needed for something else; anyway, I can live without it *)
            | 8 -> delete c
	    | 13 when enter.contents= 0 -> 
                    c.before <- [];
                    c.after <- ['Q';'u';'i';'t'];
                    enter := 5
       | 13 when enter.contents = 3 ->
                   let rg= get_string c and
                       (i, j)= xs.cpos in 
                     ( ( try
                           let pos=search rg xs.tx (j+2) in  (* using the STR library on Ocaml -> Str.search_forward (Str.regexp rg)*) 
                             glb.search_pos <- pos
                         with
                           any -> glb.search_pos <- j
                       )
                      );
                      c.before <- [];
							 c.after <- ['D';'o';'n';'e'];
                      enter := 5
       | 13 when enter.contents = 4 ->
		                    let rg= get_string c and
		                        (i, j)= xs.cpos in 
		                      ( ( try
		                            let pos= search_back rg xs.tx j in  (* using the STR library on Ocaml -> Str.search_backward (Str.regexp rg)*) 
		                              glb.search_pos <- pos
		                          with
		                            any -> glb.search_pos <- j
		                        )
		                       );
		                       c.before <- [];
		 							 c.after <- ['D';'o';'n';'e'];
                      enter := 5
       | 13 when enter.contents= 1 && id >= 0 ->
                   let newnm= get_string c in 
                     change_name pilha.(id) newnm; 
							c.before <- [];
							c.after <- ['L';'o';'a';'d';'e';'d'];
							enter := 5 
       | 13 when enter.contents= 2 && id >= 0 ->
				         let newnm= get_string c in 
				            pilha.(id) <- newnm;
								savt (pilha.(id)) xs.tx;
								c.before <- [];
								c.after <- ['W';'r';'i';'t';'t';'e';'n'];
							enter := 5
	    | 19 when id >= 0 -> savt (pilha.(id)) xs.tx; 
                    c.before <- [];
                    c.after <- ['S';'a';'v';'e';'d'];
                    enter := 5
	    | 23 -> stpYes := true;
                    enter := 2;
                    c.before <- [];
		    c.after <- ['C';'n';'t';'r';'l';'-';'W'];
                    erase_sc := true
	    | 6 when stpYes.contents ->  step c
       | 6 ->   stpYes := true; 
                     enter := 1;
                     c.before <- [];
                     c.after <- ['C';'n';'t';'r';'l';'-';'F'];
                     erase_sc := true
            | 27 -> ()    
           
            | 1 ->  print_int !ascii;
                    print_newline()
            | i when i > 31 -> insert c !ch
            | _ -> ());
         ascii := int_of_char !ch;
         sc := get_all c;
         draw_line ();
         if !enter <> 5 then (ch := Graphics.read_key());
         if erase_sc.contents then (c.before <- []; c.after <- []; erase_sc := false);
         
         
  done;
  ch := '-';
  xxx := get_string c;
  (*c.before <- []; c.after <- []*) in
  {line_input=read_string; state= (fun cs -> (sc := cs)); draw_line= draw_line};;


(*eMac like Text editor *)

(*Text editor *)
let make_edt 
	?(fn=fun s -> s) 
	?(the_text= "")
	name  (x0, y0, w, h)=
let sts= {	bkpx=0; 
		nm= name; 
(* togle to delete the char or not when using the backspace key *)
		bkerase= true; 
		ctx_save= ""; cmk_save="";
(* buffer to save the response of CTRL-T *)
		sty_stk= []; 
		insertions= false; 
		etx= {
				tx= "";
				cpos= (0,0);
				stx=0;
				sty=0; mark=0; mask= ""};
		virtual_char= 0} in
let (tw, th)= resize_char (Graphics.text_size "w") and
	rc = mkrec x0 y0 w h  in
let id= push sts.nm and
(* stack control of the sty - scrool routines *)
	push_y v= 
   	if sts.sty_stk <> [] && v= List.hd sts.sty_stk then ()
		else (sts.sty_stk <- v::sts.sty_stk)
and pop_y () = match sts.sty_stk with
		  [] -> 0
		| (x::xs) -> sts.sty_stk <- xs; x
in

(* Cursor movements *) 

(* find the end of the text *)
let last_position c =
  let rec find_last_char_non_space txt position =
    match txt.[position] with
	    ch when ch=' ' -> find_last_char_non_space txt (position- 1)
	  | _ -> (position)
  in find_last_char_non_space c.tx (String.length c.tx -1) 
in
(* reset the screen editor *)
let reset_editor c= 
	c.cpos <- (0,0); sts.sty_stk <- []; c.stx <- 0; c.sty <- 0 
in
(* swap ctx-buffer and c.tx routine *)
let ctx_swap c= 
   if sts.ctx_save= "" then ()
   else ( let ctx= sts.ctx_save in
		sts.ctx_save <- c.tx;
		c.tx <- ctx;
		reset_editor c)
in
(* find the start of the next line *)
let rec next_line position texto chars_per_line = 
  match String.length texto with
	   j when position>=j -> position;
	| _ when chars_per_line <=0 -> position+1;
	| _ when texto.[position] = '\013' -> position+1;
	| _ -> next_line (position+1) texto (chars_per_line - 1) 
in
(* Scroll the screen *)
let scroll c =
	push_y c.sty;
	c.sty <- (next_line c.sty c.tx glb.chars_per_line);
in
let scroll_back c =
	c.sty <- pop_y ();
in
(* go a step forword *)
let stp c = match c.cpos with
	  (i,j) when j >= (String.length c.tx) -> ()
(* if out or end of the string do nothing *)
	| (i,j) when c.tx.[j] = '\013' && 
	             i < (glb.lines_per_page- 1) -> c.stx <- 0; c.cpos  <-  (i+1,j+1);
(* if found newline and its not end of the line just go to next line *)
	| (i,j) when c.tx.[j] = '\013'  && 
	             i >= (glb.lines_per_page- 1) -> scroll c ; c.stx <- 0; c.cpos  <-  (i,j+1);
(* if found newline and its end of the window do a scroll *)
	| (i,j) when c.stx >= glb.chars_per_line 
	             && i < (glb.lines_per_page- 1) -> c.stx <- 0; c.cpos <- (i+1,j+1);
(* if end of line (window) and do not end of screen go next line *)
	| (i,j) when c.stx >= glb.chars_per_line 
	             && i >= (glb.lines_per_page- 1) -> scroll c; c.stx <- 0; c.cpos <- (i,j+1);
(* if end of line (window) and end of screen do scroll *)
	| (i,j) -> c.cpos <- (i , j+ 1); c.stx <- c.stx+1;
(* just go next step *)
in
(* go a step backword *)
let bkstep c = 
	match c.cpos with
	  (i,j) when j<=0 -> ()
(* if start of the string do nothing *)
	| (i,j) when c.tx.[j-1] = '\013' && i<=0 -> scroll_back c; c.cpos <- (i,j-1);
(* if find newline and first line scroll back *)
	| (i,j) when c.tx.[j-1] = '\013' && i>0 ->  c.cpos <- (i-1,j-1);
(* if find newline just go back *)
	| (i,j) when c.stx = 0 && i<=0 -> scroll_back c; c.cpos <- (i,j-1);
(* if first position and first line scroll back *)
	| (i,j) when c.stx <=0 -> c.cpos <- (i-1,j-1);
(* if first position and not first line go back *)
	| (i,j) -> c.cpos <- (i,j-1);
in
(* goto the position on the text *)
let rec goto_position c position=
  match c.cpos with
      (i, j) when position <0 || j> position -> ()
    | _ -> stp c; goto_position c position
in
(* goto the end of the document *)
let goto_end_of_doc c= reset_editor c; goto_position c (last_position c)
in
(* goto n positions back *)
let rec goto_pos_bk c pos =
	if (c.stx <=pos)  then ()
	else (bkstep c;  c.stx <-  c.stx -1; goto_pos_bk c pos)
in
(* goto first position of the line *)
let rec goto_first_pos_of_line c = 
	c.cpos <- ((fst c.cpos), (snd c.cpos - c.stx)); c.stx <- 0;
in
(* go previous line step 1 *)
let upstp c = 
	sts.bkpx <- c.stx;
	c.cpos <- ((fst c.cpos), (snd c.cpos - c.stx)); c.stx <- 0;
	bkstep c;
	sts.virtual_char <- 31; 
in
(* go previous line step 2 *)
let upstp_2 c = 
	goto_pos_bk c sts.bkpx; 
in
(* goto on a position in the line *)
let rec goto_pos c pos = match c.cpos with
	  (i, j) when j>=(String.length c.tx) -2 -> ()
	| (i, j) when c.stx>= glb.chars_per_line && pos>= glb.chars_per_line -> ();
	| (i, j) when c.stx>= glb.chars_per_line -> ()
	| (i, j) when c.stx>=pos -> ()
	| (i, j) when c.tx.[(snd c.cpos)]= '\013' -> ()
	| _ -> stp c; goto_pos c pos
in
(* goto end of the line *)
let goto_end_of_line c=
	( if c.stx = glb.lines_per_page then () );
	goto_pos c ( glb.chars_per_line );
in
let rec nxt_line_loop c=
	if c.stx=0 || (snd c.cpos)> (String.length c.tx)- 2 then ()
	else (stp c; nxt_line_loop c) 
in
(* goto first position of the next line *)
let nxt_line c = stp c; nxt_line_loop c;
in
(* go down one line on the same position *)
let dwnstp c = 
   sts.bkpx <- c.stx; nxt_line c; goto_pos c sts.bkpx;
in
(* insert a char on the text *)
let rec ins_char ch c = match c.cpos with
	   (i,j) when j> (String.length c.tx) -2 -> ()
	| (i,j)  -> 
            let x= c.tx  in
            let
                tr= (String.length x) - 2  - j and
                fr= (j+1) in
              ( String.blit x j x fr tr;
                 x.[j] <- ch; stp c)

in
let ins_txt c txt= 
   for i= 0 to String.length txt - 1 do
      ins_char txt.[i] c
   done
in
let del_char c = match c.cpos with
	(i,j) -> String.blit c.tx (j+1) c.tx j ((String.length c.tx)  - j - 1);
				c.tx.[String.length c.tx -1] <- ' '
in
(* delete j caracteres from the cursor *)
let rec del_them j c=
   if j < 1 then ()
   else ( del_char c; del_them (j-1) c ) 
in
(*  delete n caracteres back the cursor *)
let rec del_back n c=
   if n < 0 then ()
   else (bkstep c; del_char c; del_back (n-1) c)
in
(* delete from mark to cursor *)
let del_txt c = match c.cpos with
		(i, j) when j > c.mark -> ins_char (char_of_int 13) c ;del_back (j - c.mark) c
	 | (i, j) -> del_them (c.mark - j) c
in
(* Reset the cursor on the display *)
let reset_display c = c.cpos <- (0, c.sty); c.stx <- 0;
in
(* put the cursor on the mouse position *)
let mouse_point c = 
  let rec loop i =
	if i<=0 then ()
	else (dwnstp c; loop (i-1) ) in
  let (x,y)= glb.mouse_position in
  let (cx, cy)= (min glb.chars_per_line ((x-rc.x-3+caractere_space)/tw), min glb.lines_per_page ((rc.y+rc.h-3+th-y)/th)) in
  reset_display c;
  loop (cy-1);
  goto_first_pos_of_line c ; goto_pos c cx;
(* put the cursor at the cx position *)
in
(* find the position of the char after the cursor *)
let find_char char c=
 let rec loop j=
    if (j>String.length c.tx -1) || (c.tx.[j]=char) then j
    else loop (j+1) in
 loop (snd c.cpos)
in
let rec find_to j n str= match n with
       t when t < (String.length str) - 2 && str.[t] = '\009' && str.[t+1]= ' ' ->
                t - j
     | t when t < (String.length str) - 2 -> find_to j (n+1) str
     | t -> (String.length str) - 2  - j 
in
let contour= ref (Graphics.red)
(*and edTxt= ref {tx= "";  cpos= (0,0); stx=0; sty=0} *) in

let set_state cpos gcor i=
   sts.etx <- cpos;
   contour := (get_color gcor);
   if i then ( sts.etx.sty <- 0; sts.etx.stx <- 0; sts.sty_stk <- [] ) in
let (size_x, size_y)= (Graphics.size_x(), Graphics.size_y()) in
let ctrl_x= make_line_input 20
               (mkrec 25 (rc.y - 2*th) (size_x - 50) (2*th)) 
in
let process_esc k c= match c.cpos with
	  (i, j) when k= ' ' -> c.mark <- j
	| (i, j) when k= 'w' && j > c.mark -> 
	             push_circular ( String.sub c.tx  c.mark (j- c.mark))
	| (i, j) when k= 'w' && c.mark > j -> 
	             push_circular ( String.sub c.tx  j (c.mark - j))
	| (i, j) when k= 'a' -> reset_editor c 
	| (i, j) when k= 'e' -> goto_end_of_doc c
	| (i, j) when k= 'y' && glb.mrk>=0 && glb.mrk < j ->
	             c.mark <- glb.mrk; del_txt c; circula(); ins_txt c (getop ())
	| (i, j) -> () 
(* the position of the tag box *)
and tg= tag id rc 
  (* Control a virtual keypress *)
and   pop_virtualchar () =
    let c= sts.virtual_char in ( sts.virtual_char <- 0; c )
in
  (* Control of mini buttons in the window editor *)
  (* position of the mini buttons *)
let   mini_rc  = {x= rc.x+rc.w-15; y= rc.y+rc.h-15; w= 15; h= 15 }
and scroll_up_bt = {x= rc.x+rc.w-15; y= rc.y+rc.h-40; w= 15; h= 15 }
and scroll_down_bt = {x= rc.x+rc.w-15; y= rc.y; w= 15; h= 15 }
in
(* draw of tre mini button *)
let draw_mini_buttons () =
    Graphics.draw_rect mini_rc.x mini_rc.y mini_rc.w mini_rc.h;
    Graphics.draw_rect scroll_up_bt.x scroll_up_bt.y scroll_up_bt.w scroll_up_bt.h;
    Graphics.draw_rect scroll_down_bt.x scroll_down_bt.y scroll_down_bt.w scroll_down_bt.h;
    Graphics.moveto (mini_rc.x + 3) mini_rc.y; Graphics.draw_char 'X'  ;
    Graphics.moveto (scroll_up_bt.x+1) (scroll_up_bt.y+1);
    Graphics.lineto (scroll_up_bt.x+scroll_up_bt.w/2) (scroll_up_bt.y+scroll_up_bt.h/2+4);
    Graphics.lineto (scroll_up_bt.x+scroll_up_bt.w-1) (scroll_up_bt.y);
    Graphics.moveto (scroll_down_bt.x) (scroll_down_bt.y+scroll_down_bt.h);
    Graphics.lineto (scroll_down_bt.x+scroll_down_bt.w/2+1) (scroll_down_bt.y+scroll_down_bt.h/2-4);
    Graphics.lineto (scroll_down_bt.x+scroll_down_bt.w) (scroll_down_bt.y+scroll_down_bt.h);in
let 
  draw_editor ()=
   Graphics.auto_synchronize false; 
   let tg= tag id rc  in
   draw_box contour.contents 2 tg; 
   Graphics.set_color background_edt; 
   Graphics.fill_rect rc.x rc.y rc.w rc.h; 
   Graphics.set_color contour.contents; 
   Graphics.draw_rect rc.x rc.y rc.w rc.h;
   draw_mini_buttons ();
   Graphics.set_color Graphics.black;
   draw_string (tg.x+6) (tg.y+ (tg.h-th)/2) (pilha.(id)); 
   drawText  (rc.x+3)  (rc.y + rc.h - th- 3) sts.etx; 
   ctrl_x.draw_line ();
   Graphics.auto_synchronize true
and
 in_muride ()=
   if  mouse_in_window (Graphics.mouse_pos()) tg  then
           (sts.insertions <- true; char_of_int 21) 
   else   (sts.insertions <- false; char_of_int 0) 
and
   in_char ()=
  if sts.virtual_char <> 0 then char_of_int (pop_virtualchar () )
(* if there are any key on the virtual keybord then return the top one *)
  else begin
    let ev= Graphics.wait_next_event [Graphics.Button_down;  Graphics.Key_pressed] in
(* wait for a key pressed or a mouse button down *)
      glb.mouse_position <- Graphics.mouse_pos();
(* storage the position of the mouse *)
	match ev with 
	   event when event.Graphics.keypressed -> ev.Graphics.key
(* if a key is pressed then return this key *)
	| event when (mouse_in_window glb.mouse_position tg) -> sts.insertions <- true; char_of_int 0
(* if click on the tag active the editor id *)
	| event when (mouse_in_window glb.mouse_position scroll_up_bt) -> sts.insertions <- true; char_of_int 21
(* if click on scroll down button do it *)
	| event when (mouse_in_window glb.mouse_position scroll_down_bt) -> sts.insertions <- true; char_of_int 09
(* if click on scroll up button do it *)
	| event when (mouse_in_window glb.mouse_position  mini_rc) -> char_of_int 20
(* if click on the mini-button then return a ascii 20 *)
	| event when (mouse_in_window glb.mouse_position rc) -> sts.insertions <- true; char_of_int 28
(* if mouse in the editor window then move the cursor to it position on keypressed ascii 28 *)
	| _ -> sts.insertions <- false; char_of_int 0
  end
in
 set_state {tx= "";  cpos= (0,0); stx=0; sty=0; mark=0; 
             mask= ""} (CLR Graphics.red) true;
ctrl_x.state ['-'];
draw_editor ();

(* This is the process event! *) 
let text_input ()= 
let lines = if oldStuff.(id) = "" then the_text
           else oldStuff.(id) in
let 
  c=  createText (if lines="" then (String.make 500 ' ') else lines) and
  ch= ref (char_of_int 0) in
  (  
     c.stx <- 0;
     c.sty <- 0;
     ch :=in_muride (); 
     while sts.insertions do
         if !ch<> char_of_int 21 
         then ch := in_char () else (ch:= char_of_int 0); 
         if !ch <> (char_of_int 27) then glb.mrk <- (-1) else ();
      (*mainkeys*)   
      ( match (int_of_char !ch) with
	   01 (*Ctrl-a Goto first position of the line *) -> goto_first_pos_of_line c
	| 02 (*Ctrl-b - Backstep  *) ->  bkstep c
	| 03 (*Ctrl-c *) -> ()
	| 04 (*Ctrl-d Backspace *) -> del_char c 
	| 05 (*Crtl-e  goto end of line*) -> goto_end_of_line c
	| 06 (*Crtl-f *)-> stp c
	| 07 (*Ctrl-g *) -> ()
	| 08 (*Ctrl-h & Backspace *) -> if (snd c.cpos)>0 then 
			( bkstep c ; if sts.bkerase then del_char c) 
	| 09 (*Ctrl-i *) -> scroll c;sts.virtual_char <- 14; 
	| 10 (*Ctrl-j insert a line break  *) -> ins_char (char_of_int 13) c (* standard behavior *)
															(* nxt_line c *)  			(* non standard behavior *)
	| 11 (*Ctrl-k kills to the end of line *) ->  
			c.mark <- find_char '\013' c;
			if c.mark= (snd c.cpos) then del_char c
			else (process_esc 'w' c; del_txt c)
	| 12 (*Ctrl-m *) -> ()
	| 13 (*ENTER *) -> ins_char (char_of_int 13) c 
	| 14 (*Ctrl-n Move to next line *) -> dwnstp c;

	| 15 (*Ctrl-o *) ->ins_char (char_of_int 13) c; bkstep c
	| 16 (*Ctrl-P Move to previous Line Step-1*) -> upstp  c
	| 31 (*Ctrl-P Move to previous Line Step-2*) -> upstp_2 c
	| 17 (*Ctrl-q Toggle between bkstep and bkerase *) -> sts.bkerase <- not sts.bkerase
	| 18 (*Ctrl-r search text backward *) -> ctrl_x.line_input 2 id c;reset_editor c; goto_position c (glb.search_pos -2);
	| 19 (*Ctrl-s search text forward *) -> ctrl_x.line_input 1 id c;reset_editor c; goto_position c (glb.search_pos -2)
	| 20 (*Ctrl-t: Applica funo *) -> 
		sts.ctx_save <- c.tx; 
		reset_editor c;
		let sz= String.length c.tx and
		     new_tx= fn c.tx in
		let new_sz= String.length new_tx in
		if new_sz > sz then ()
			else ( c.tx <- new_tx ^ (String.make (sz- new_sz) ' '))
	| 21 (*Ctrl-u *) -> scroll_back c;sts.virtual_char <- 16; 
	| 22 (*Ctrl-v *) -> ctx_swap c
	| 23 (*Ctrl-w Erase marked area *) ->  process_esc 'w' c;del_txt c 
	| 24 (*Ctrl-x Execute function *) ->  ctrl_x.state ['C';'n';'t';'r';'l';'-';'X'];
	                                        ctrl_x.draw_line ();
	                                        oldStuff.(id) <- c.tx;
	                                        ctrl_x.line_input 0 id c;
	                                        c.tx <- oldStuff.(id);
	                                        reset_editor c;
	                                        draw_editor()

	| 25 (*Ctrl-y paste the buffer *) when getop () <> "" -> 
	            let (i, j)= c.cpos in
	               glb.mrk <- j;
	               ins_txt c (getop ())
	| 26 (*Ctrl-z *) -> sts.insertions <- false
	| 27 (* Esc *) -> process_esc (Graphics.read_key()) c
	| 28          -> mouse_point c; 
	| x when ch.contents >= ' '  (* Insert char *) -> ins_char ch.contents c  (**)
	| _ -> ());

        set_state c (CLR Graphics.blue) false;
        draw_editor();
        
     done;
     if !contour= Graphics.blue then (set_state c (CLR Graphics.red) true; draw_editor())
     else ();
     sts.insertions <- false;
     
   
     c.stx <- 0;
     c.sty <- 0;
     sts.sty_stk <- []
  ) in oldStuff.(id) <- the_text;
       eds.(id) <- {process_event=text_input; st= set_state; draw_self= draw_editor};
       glb.refresh_list <- eds.(id)::glb.refresh_list;
       eds.(id)
;;
             
let mk_edt ?(txt= String.make 400 ' ')  
		?(fn= fun s -> s) 
		nm=
		  let (cw, ch)= resize_char (Graphics.text_size "w") in
		  let (size_x, size_y)= (Graphics.size_x(), Graphics.size_y()) in
		  let rc= (25, size_y - (winHeight+6)*ch, size_x - 50, 100) in
		glb.chars_per_line <- (size_x - 100) / cw;  
   make_edt  ~the_text:txt  ~fn:fn nm  rc
;;

let make_button fn txt (x0, y0)=
let (cw, ch)= Graphics.text_size "w" in
let (w, h)= (12*cw, 2*ch) in
let (tw, th)= Graphics.text_size txt and 
    rc = mkrec x0 y0 w h  and
    pressed = ref false and
    stat= ref {cpos= (0,0);  tx=txt; stx=0; sty=0; mark=0; mask= ""} and
    contour= ref Graphics.red in
let draw ()=
  Graphics.auto_synchronize false;
  draw_box !contour 3 rc;
  Graphics.set_color Graphics.black;
  draw_string (x0+(w-tw)/2) (y0+(h-th)/2) !stat.tx ;
  Graphics.auto_synchronize true 
and set_state c gcor i= contour:= (get_color gcor) in
  draw ();
  let button_input ()= 
     if (click_in_window rc) && not pressed.contents  then (fn(); pressed := true; 
                               set_state !stat (CLR Graphics.blue) false; draw());
     if not(Graphics.button_down()) && pressed.contents then  (pressed := false;  
                                          set_state !stat (CLR Graphics.red) false; draw ())
      in 
          let theButton= {process_event=button_input; st= set_state; draw_self= draw}
          in ( glb.refresh_list <- theButton::glb.refresh_list;
               theButton)
;;

let refresh ()=
  let rec loop s= match s with
        [] -> ()
      | v::vs -> v.st {cpos= (0,0);  tx= ""; stx=0; sty=0; mark=0; mask= "" } 
                        (CLR Graphics.red) true; v.draw_self(); loop vs
  in
     ignore(Graphics.wait_next_event[Graphics.Mouse_motion; Graphics.Button_down]);
    
    let (sx,sy)= (Graphics.size_x(), Graphics.size_y()) in
     if last_sz.contents = (0,0) then
        (  Graphics.auto_synchronize false;   
           set_siz();
           loop glb.refresh_list;
           Graphics.auto_synchronize true )
     else ( if (sx,sy) <> last_sz.contents then 
              ( Graphics.auto_synchronize false;  
                Graphics.set_color background_mask;
                Graphics.fill_rect 0 0 sx sy; 
                Graphics.set_color Graphics.black; 
                set_siz();
                loop glb.refresh_list;
                Graphics.auto_synchronize true 
              )
           )
;;

(* defaut Exit Button *)
let exit_button ()= 
let (cw, ch)= Graphics.text_size "w" in
let (w, h)= (12*cw, 2*ch) in
let (size_x, size_y)= (Graphics.size_x(), Graphics.size_y()) in
let rc= (size_x - 25 - w, size_y-(winHeight+8)*ch - h -10) in
    make_button (fun () -> Graphics.close_graph(); exit 0) "Exit" rc
;;

let button i fs nm=
let (cw, ch)= Graphics.text_size "w" in
let (w, h)= (15*cw, 2*ch) in
let (size_x, size_y)= (Graphics.size_x(), Graphics.size_y()) in
let rc= (50+ i* w, size_y-(winHeight+8)*ch - h -10) in
    make_button fs nm rc
;;

let next_button_row i fs nm=
let (cw, ch)= Graphics.text_size "w" in
let (w, h)= (15*cw, 2*ch) in
let (size_x, size_y)= (Graphics.size_x(), Graphics.size_y()) in
let rc= (50+ i* w, size_y-(winHeight+8)*ch - 2*h -10) in
    make_button fs nm rc;;

let get_fields msk txt= 
   match String.length txt with
       sz when sz <> (String.length msk) -> txt
    |  sz -> let c= String.make sz ' ' in
                 for i= 0 to sz - 1 do
                    if msk.[i]= '#' then c.[i] <- if txt.[i]='#' then ' ' else txt.[i]
                    else c.[i] <- ' '
                 done; c;;
          
(* Mask Editor *)
let make_entry 
	?(fn=fun s -> s) 
	?(pp= get_fields)
	?(the_text= "")
	(*?(mask=fun x  -> true) *)
	name (x0, y0, w, h)=
let sts= { 	bkpx=0; 
		nm= name; 
(* togle to delete the char or not when using the backspace key *)
		bkerase= true; 
		ctx_save= "";
		cmk_save="";
		sty_stk= []; 
		insertions= false;
		etx= {	tx= "";
				cpos= (0,0); 
				stx=0; 
				sty=0; mark=0; mask=""}; 
		virtual_char= 0} and
		ed_lines= the_text  in
let (tw, th)= resize_char (Graphics.text_size "w") and
	rc = mkrec x0 y0 w h in
let id= push sts.nm and
(* stack control of the sty - scrool routines *)
	push_y v= 
   	if sts.sty_stk <> [] && v= List.hd sts.sty_stk then ()
		else (sts.sty_stk <- v::sts.sty_stk)
and pop_y () = match sts.sty_stk with
		  [] -> 0
		| (x::xs) -> sts.sty_stk <- xs; x
in

(* Cursor movements *) 

(* find the end of the text *)
let last_position c =
  let rec find_last_char_non_space txt position =
    match txt.[position] with
	    ch when ch=' ' -> find_last_char_non_space txt (position- 1)
	  | _ -> (position)
  in find_last_char_non_space c.tx (String.length c.tx -1) 
in
(* reset the screen editor *)
let reset_editor c= 
	c.cpos <- (0,0); sts.sty_stk <- []; c.stx <- 0; c.sty <- 0 
in
(* swap ctx-buffer and c.tx routine *)
let ctx_swap c= 
   if sts.ctx_save= "" then ()
   else ( let ctx= sts.ctx_save and
   		 cmk= sts.cmk_save in
		sts.ctx_save <- c.tx;
		sts.cmk_save <- c.mask;
		c.tx <- ctx;
		c.mask <- cmk;
		reset_editor c)
in
(* find the start of the next line *)
let rec next_line position texto chars_per_line = 
  match String.length texto with
	   j when position>=j -> position;
	| _ when chars_per_line <=0 -> position+1;
	| _ when texto.[position] = '\013' -> position+1;
	| _ -> next_line (position+1) texto (chars_per_line - 1) 
in
(* Scroll the screen *)
let scroll c =
	push_y c.sty;
	c.sty <- (next_line c.sty c.tx glb.chars_per_line);
in
let scroll_back c =
	c.sty <- pop_y ();
in
(* go a step forword *)
let stp c = match c.cpos with
	  (i,j) when j >= (String.length c.tx) -> ()
(* if out or end of the string do nothing *)
	| (i,j) when c.tx.[j] = '\013' && 
	             i < (glb.lines_per_page- 1) -> c.stx <- 0; c.cpos  <-  (i+1,j+1);
(* if found newline and its not end of the line just go to next line *)
	| (i,j) when c.tx.[j] = '\013'  && 
	             i >= (glb.lines_per_page- 1) -> scroll c ; c.stx <- 0; c.cpos  <-  (i,j+1);
(* if found newline and its end of the window do a scroll *)
	| (i,j) when c.stx >= glb.chars_per_line 
	             && i < (glb.lines_per_page- 1) -> c.stx <- 0; c.cpos <- (i+1,j+1);
(* if end of line (window) and do not end of screen go next line *)
	| (i,j) when c.stx >= glb.chars_per_line 
	             && i >= (glb.lines_per_page- 1) -> scroll c; c.stx <- 0; c.cpos <- (i,j+1);
(* if end of line (window) and end of screen do scroll *)
	| (i,j) -> c.cpos <- (i , j+ 1); c.stx <- c.stx+1;
(* just go next step *)
in
(* go a step backword *)
let bkstep c = 
	match c.cpos with
	  (i,j) when j<=0 -> ()
(* if start of the string do nothing *)
	| (i,j) when c.tx.[j-1] = '\013' && i<=0 -> c.sty <- pop_y (); c.cpos <- (0,j-1);
(* if find newline and first line scroll back *)
	| (i,j) when c.tx.[j-1] = '\013' && i>0 ->  c.cpos <- (i-1,j-1);
(* if find newline just go back *)
	| (i,j) when c.stx = 0 && i<=0 -> c.sty <- pop_y (); c.cpos <- (0,j-1);
(* if first position and first line scroll back *)
	| (i,j) when c.stx <=0 -> c.cpos <- (i-1,j-1);
(* if first position and not first line go back *)
	| (i,j) -> c.cpos <- (i,j-1);
in
(* goto the position on the text *)
let rec goto_position c position=
  match c.cpos with
      (i, j) when j> position -> ()
    | _ -> stp c; goto_position c position
in
(* goto the end of the document *)
let goto_end_of_doc c= reset_editor c; goto_position c (last_position c)
in
(* goto n positions back *)
let rec goto_pos_bk c pos =
	if (c.stx <=pos)  then ()
	else (bkstep c;  c.stx <-  c.stx -1; goto_pos_bk c pos)
in
(* goto first position of the line *)
let rec goto_first_pos_of_line c = 
	c.cpos <- ((fst c.cpos), (snd c.cpos - c.stx)); c.stx <- 0;
in
(* go previous line step 1 *)
let upstp c = 
	sts.bkpx <- c.stx;
	c.cpos <- ((fst c.cpos), (snd c.cpos - c.stx)); c.stx <- 0;
	bkstep c;
	sts.virtual_char <- 31; 
in
(* go previous line step 2 *)
let upstp_2 c = 
	goto_pos_bk c sts.bkpx; 
in
(* goto on a position in the line *)
let rec goto_pos c pos = match c.cpos with
	  (i, j) when j>=(String.length c.tx) -2 -> ()
	| (i, j) when c.stx>= glb.chars_per_line && pos>= glb.chars_per_line -> ();
	| (i, j) when c.stx>= glb.chars_per_line -> ()
	| (i, j) when c.stx>=pos -> ()
	| (i, j) when c.tx.[(snd c.cpos)]= '\013' -> ()
	| _ -> stp c; goto_pos c pos
in
(* goto end of the line *)
let goto_end_of_line c=
	( if c.stx = glb.lines_per_page then () );
	goto_pos c ( glb.chars_per_line );
in
let rec nxt_line_loop c=
	if c.stx=0 || (snd c.cpos)> (String.length c.tx)- 2 then ()
	else (stp c; nxt_line_loop c) 
in
(* goto first position of the next line *)
let nxt_line c = stp c; nxt_line_loop c;
in
(* go down one line on the same position *)
let dwnstp c = 
   sts.bkpx <- c.stx; nxt_line c; goto_pos c sts.bkpx;
in
(* insert a char on the text if its posible *)
let rec over_char ch c = match c.cpos with
   |  (i,j) when  j < (String.length c.tx) - 1  -> 
            let x= c.tx  in
              ( if (mask_fun c.mask) j then x.[j] <- ch; stp c)
   |  _ -> ()

in
let ins_txt c txt= 
   for i= 0 to String.length txt - 1 do
      over_char txt.[i] c
   done
in
let del_char c = match c.cpos with
	(i,j) -> if (mask_fun c.mask) j then c.tx.[j] <- '#'
in
(* delete j caracteres from the cursor *)
let rec del_them j c=
   if j < 1 then ()
   else ( del_char c; del_them (j-1) c ) 
in
(* delete from mark to cursor *)
let del_txt c = match c.cpos with
		(i, j) when j > c.mark -> ()
	 | (i, j) -> del_them (c.mark - j) c
in
let rec find_to j n str= match n with
       t when t < (String.length str) - 2 && str.[t] = '\009' && str.[t+1]= ' ' ->
                t - j
     | t when t < (String.length str) - 2 -> find_to j (n+1) str
     | t -> (String.length str) - 2  - j 
in
(* backstep on a field text *)
let bkstep_field c =
  let rec bkstep_field_loop c =
	match c.tx.[snd c.cpos] with
		  x when x='#' || x='@' -> bkstep c;  bkstep_field_loop c; 
		| _ -> stp c
	in match (c.tx.[ snd c.cpos]) with
		x when x='#' || x='@'  -> bkstep_field_loop c
		| x when (int_of_char x)=13 || x = ' ' -> bkstep c;  bkstep_field_loop c
		| _ -> ()
in

let contour= ref (Graphics.red)
(*and edTxt= ref {tx= "";  cpos= (0,0); stx=0; sty=0 } *) in
(* Reset the cursor on the display *)
let reset_display c = c.cpos <- (0, c.sty); c.stx <- 0;
in
(* put the cursor on the mouse position *)
let mouse_point c = 
  let rec loop i =
	if i<=0 then ()
	else (dwnstp c; loop (i-1) ) in
  let (x,y)= glb.mouse_position in
  let (cx, cy)= (min glb.chars_per_line ((x-rc.x-3+caractere_space)/tw), min glb.lines_per_page ((rc.y+rc.h-3+th-y)/th)) in
  reset_display c;
  loop (cy-1);
  goto_first_pos_of_line c ; goto_pos c cx;
(* put the cursor at the cx position *)
in
let set_state cpos gcor i=
   sts.etx <- cpos;
   contour := (get_color gcor);
   if i then ( sts.etx.sty <- 0; sts.etx.stx <- 0; sts.sty_stk <- [] ) in
let (size_x, size_y)= (Graphics.size_x(), Graphics.size_y()) in
let ctrl_x= make_line_input 20
               (mkrec 25 (rc.y - 2*th) (size_x - 50) (2*th)) 
in
let process_esc k c= match c.cpos with
	  (i, j) when k= ' ' -> c.mark <- j
	| (i, j) when k= 'w' -> push_circular ( String.sub c.tx (min c.mark j) 
	                                          (abs (j- c.mark)))
	| (i, j) when k= 'a' -> reset_editor c 
	| (i, j) when k= 'e' -> goto_end_of_doc c
	| (i, j) -> () 
and tg= tag id rc 
  (* Control a virtual keypress *)
and pop_virtualchar () =
    let c= sts.virtual_char in ( sts.virtual_char <- 0; c )
in
  (* Control of mini buttons in the window editor *)
  (* position of the mini buttons *)
let   mini_rc  = {x= rc.x+rc.w-15; y= rc.y+rc.h-15; w= 15; h= 15 }
and scroll_up_bt = {x= rc.x+rc.w-15; y= rc.y+rc.h-40; w= 15; h= 15 }
and scroll_down_bt = {x= rc.x+rc.w-15; y= rc.y; w= 15; h= 15 }
in
(* draw of tre mini button *)
let draw_mini_buttons () =
    Graphics.draw_rect mini_rc.x mini_rc.y mini_rc.w mini_rc.h;
    Graphics.draw_rect scroll_up_bt.x scroll_up_bt.y scroll_up_bt.w scroll_up_bt.h;
    Graphics.draw_rect scroll_down_bt.x scroll_down_bt.y scroll_down_bt.w scroll_down_bt.h;
    Graphics.moveto (mini_rc.x + 3) mini_rc.y; Graphics.draw_char 'X'  ;
    Graphics.moveto (scroll_up_bt.x+1) (scroll_up_bt.y+1);
    Graphics.lineto (scroll_up_bt.x+scroll_up_bt.w/2) (scroll_up_bt.y+scroll_up_bt.h/2+4);
    Graphics.lineto (scroll_up_bt.x+scroll_up_bt.w-1) (scroll_up_bt.y);
    Graphics.moveto (scroll_down_bt.x) (scroll_down_bt.y+scroll_down_bt.h);
    Graphics.lineto (scroll_down_bt.x+scroll_down_bt.w/2+1) (scroll_down_bt.y+scroll_down_bt.h/2-4);
    Graphics.lineto (scroll_down_bt.x+scroll_down_bt.w) (scroll_down_bt.y+scroll_down_bt.h);in

let 
  draw_editor ()=
   Graphics.auto_synchronize false; 
   let tg= tag id rc  in
   draw_box contour.contents 2 tg; 
   Graphics.set_color background_mask; 
   Graphics.fill_rect rc.x rc.y rc.w rc.h; 
   Graphics.set_color contour.contents; 
   Graphics.draw_rect rc.x rc.y rc.w rc.h;
   draw_mini_buttons ();
   Graphics.set_color Graphics.black;
   draw_string (tg.x+6) (tg.y+ (tg.h-th)/2) (pilha.(id)); 
   drawMask  (rc.x+3)  (rc.y + rc.h - th- 3) sts.etx; 
   ctrl_x.draw_line ();
   Graphics.auto_synchronize true
and
  in_muride () = 
     if  mouse_in_window (Graphics.mouse_pos()) tg  then
           (sts.insertions <- true; char_of_int 21) 
     else   (sts.insertions <- false; char_of_int 0) 
 and 
   in_char ()=
  if sts.virtual_char <> 0 then char_of_int (pop_virtualchar () )
(* if there are any key on the virtual keybord then return the top one *)
  else begin
    let ev= Graphics.wait_next_event [Graphics.Button_down;  Graphics.Key_pressed] in
(* wait for a key pressed or a mouse button down *)
      glb.mouse_position <- Graphics.mouse_pos();
(* storage the position of the mouse *)
	match ev with 
	   event when event.Graphics.keypressed -> ev.Graphics.key
(* if a key is pressed then return this key *)
	| event when (mouse_in_window glb.mouse_position tg) -> sts.insertions <- true; char_of_int 0
(* if click on the tag active the editor id *)
	| event when (mouse_in_window glb.mouse_position scroll_up_bt) -> sts.insertions <- true; char_of_int 21
(* if click on scroll down button do it *)
	| event when (mouse_in_window glb.mouse_position scroll_down_bt) -> sts.insertions <- true; char_of_int 09
(* if click on scroll up button do it *)
	| event when (mouse_in_window glb.mouse_position  mini_rc) -> char_of_int 20
(* if click on the mini-button then return a ascii 20 *)
	| event when (mouse_in_window glb.mouse_position rc) -> sts.insertions <- true; char_of_int 28
(* if mouse in the editor winodw then move the cursor to it position on keypressed ascii 28 *)
	| _ -> sts.insertions <- false; char_of_int 0
  end
in
 set_state {tx= "";  cpos= (0,0); stx=0; sty=0; mark=0; mask= ""} (CLR Graphics.red) true;
 
ctrl_x.state ['-'];
draw_editor ();

(* This is the process event! *) 
let text_input ()= 
let lines = if oldStuff.(id) = "" then the_text
           else oldStuff.(id) and 
     mascara= if mskStuff.(id)= "" then  (String.copy the_text)
                    else mskStuff.(id)  in
let 
  c=  createMaskText lines (*if lines="" then (String.make 500 ' ') else lines*)  
                                  mascara and
  ch= ref (char_of_int 0) in
  (  
     c.stx <- 0;
     c.sty <- 0;
     ch :=in_muride (); 
     while sts.insertions do
         if !ch<> char_of_int 21 
         then ch := in_char () else (ch:= char_of_int 0); 
          
      (*mainkeys*)   
      ( match (int_of_char !ch) with
	   01 (*Ctrl-a Goto first position of the line *) -> goto_first_pos_of_line c
	| 02 (*Ctrl-b - Backstep  *) ->  bkstep c
	| 03 (*Ctrl-c *) -> ()
	| 04 (*Ctrl-d Backspace *) -> del_char c 
	| 05 (*Crtl-e  goto end of line*) -> goto_end_of_line c
	| 06 (*Crtl-f *)-> stp c
	| 07 (*Ctrl-g *) -> ()
	| 08 (*Ctrl-h & Backspace *) -> if (snd c.cpos)>0 then 
			( bkstep c; if sts.bkerase then del_char c)
	| 09 (*Ctrl-i *) -> scroll c;sts.virtual_char <- 14; 
	| 10 (*Ctrl-j insert a line break  *) -> over_char '\n' c
	| 11 (*Ctrl-l *) -> ()
	| 12 (*Ctrl-m *) -> ()
	| 13 (*ENTER *) -> nxt_line c
	| 14 (*Ctrl-n Move to next line *) -> dwnstp c;

	| 15 (*Ctrl-o *) -> ()
	| 16 (*Ctrl-P Move to previous Line Step-1*) ->  upstp  c
	| 31 (*Ctrl-P Move to previous Line Step-2*) -> upstp_2 c
	| 17 (*Ctrl-q Toggle between bkstep and bkerase *) -> sts.bkerase <- not sts.bkerase
	| 18 (*Ctrl-r search text backward *) -> ctrl_x.line_input 2 id c;
	                                      reset_editor c; 
	                                      goto_position c (glb.search_pos -2);
	| 19 (*Ctrl-s search text forward    *) -> ctrl_x.line_input 1 id c;
	                                           reset_editor c;
	                                           goto_position c (glb.search_pos -2)
	| 20 (*Ctrl-t: Applica funo *) -> 
		sts.ctx_save <- c.tx; 
		sts.cmk_save <- c.mask;
		reset_editor c;
		c.tx <- fn (pp c.mask c.tx) 
	| 21 (*Ctrl-u *) -> scroll_back c;sts.virtual_char <- 16; 
	| 22 (*Ctrl-v *) -> ctx_swap c
	| 23 (*Ctrl-w Erase marked area *) ->  process_esc 'w' c;del_txt c 
	| 24 (*Ctrl-x Execute function *) -> ctrl_x.state ['C';'n';'t';'r';'l';'-';'X'];
	                                        ctrl_x.draw_line ();
	                                        oldStuff.(id) <- c.tx;
	                                        ctrl_x.line_input 0 id c;
	                                        c.tx <- oldStuff.(id);
	                                       (* oldStuff.(id) <- c.mask; *)
	                                        reset_editor c;
	                                        draw_editor()

	| 25 when getop ()<>"" -> ins_txt c (getop ())
	| 26 (*Ctrl-z *) -> sts.insertions <- false
	| 27 (* Esc *) -> process_esc (Graphics.read_key()) c
	| 28          -> mouse_point c;bkstep_field c;
	| x when ch.contents >= ' '  (* Insert char *) -> over_char ch.contents c  (**)
	| _ -> ());

        set_state c (CLR Graphics.blue) false;
        draw_editor();
     done;
          if !contour= Graphics.blue then (set_state c (CLR Graphics.red) true; draw_editor())
     else ();
     sts.insertions <- false;
     c.stx <- 0;
     c.sty <- 0;
     sts.sty_stk <- []
  ) in oldStuff.(id) <- ed_lines;
       mskStuff.(id) <- String.copy ed_lines;
       eds.(id) <- {process_event=text_input; st= set_state; draw_self= draw_editor};
       glb.refresh_list <- eds.(id)::glb.refresh_list;
       eds.(id)
;;

let mk_entry 	?(txt="tot")
            	?(fn= fun s -> s)
            	?(pp= get_fields) nm  =
let (cw, ch)= resize_char (Graphics.text_size "w") in
let (size_x, size_y)= (Graphics.size_x(), Graphics.size_y()) in
let rc= (25, size_y - (winHeight+6)*ch, size_x - 50, 100) in
    glb.chars_per_line <- (size_x - 100) / cw;
   make_entry  ~the_text:(String.copy txt)  ~fn:fn ~pp:pp nm  rc;;

let the_mask xs= String.concat "\013" xs;;


(* read a file with the image in marshall format *)
let rdMsh file_nm=
   let chan= open_in_bin file_nm in
   let m= Marshal.from_channel chan in
   let sm= Array.length m.(0) and sn= Array.length m in
      (sm, sn, m)
;;

let make_gbutton img1 img2 fn txt (x0, y0)=
let (w, h, mat1)= rdMsh img1 and
    (_, _, mat2)= rdMsh img2 in
let image1= Graphics.make_image mat1 and
    image2= Graphics.make_image mat2 and 
    (char_w, char_h)= Graphics.text_size txt in
let 
    rc = mkrec x0 y0 w (h+char_h)  and
    pressed = ref false and
    stat= ref {cpos= (0,0);  tx=txt; stx=0; sty=0; mark=0; mask= "" } and
    contour= ref image1 in
let draw ()=
  Graphics.auto_synchronize false;
  Graphics.draw_image !contour rc.x (rc.y+char_h);
  draw_string (rc.x+(w-char_w)/2) rc.y txt;
  Graphics.draw_rect rc.x rc.y rc.w rc.h;
  
  Graphics.auto_synchronize true 
and set_state c gcor i= 
         contour:= (get_img image1 gcor)  in
  draw ();
  let button_input ()= 
     if (click_in_window rc) && not pressed.contents  then (fn(); pressed := true; 
                               set_state !stat (IMG image2) false; draw());
     if not(Graphics.button_down()) && pressed.contents then  (pressed := false;  
                                          set_state !stat (IMG image1) false; draw ())
      in 
          let theButton= {process_event=button_input; st= set_state; draw_self= draw}
          in ( glb.refresh_list <- theButton::glb.refresh_list;
               theButton)
;;

let gbutton i img1 img2 fs nm=
let (cw, ch)= Graphics.text_size "w" in
let (w, h)= (12*cw, 2*ch) in
let (size_x, size_y)= (Graphics.size_x(), Graphics.size_y()) in
let rc= (25+ i* w, size_y-(winHeight+8)*ch - 2*h -10) in
    make_gbutton img1 img2 fs nm rc
;;

(* returns numbers in float format in a string*)
let rec rdThem s=
  try
    match Stream.next s with
      | Genlex.Int x -> (float_of_int x)::(rdThem s)
      | Genlex.Float x -> x::(rdThem s)
      | _  -> rdThem s
  with
     _ -> []
;;

(* get a list of float using ocamlex *)
let getFloatList s=
  let lex= Genlex.make_lexer [] in
    let strm= lex (Stream.of_string s) in
         rdThem strm
;;
(* returns numbers in float format in a string*)
let rec rdstr s=
  try
    match Stream.next s with
      | Genlex.Int x -> (string_of_int x)::(rdstr s)
      | Genlex.Float x -> (string_of_float x)::(rdstr s)
      | Genlex.Ident x  -> x::(rdstr s)
      | Genlex.Kwd x -> x::(rdstr s)
      | Genlex.String x -> x::(rdstr s)
      | Genlex.Char x -> (String.make 1 x)::(rdstr s)
  with
     _ -> []
;;

(* get a list of float using ocamlex *)
let getStringList s=
  let lex= Genlex.make_lexer ["."; ",";";";"?";"!";"(";")";"+";"*";"-";"/"] in
    let strm= lex (Stream.of_string s) in
         rdstr strm
;;

let start_mask x=
  String.length x > 1 && x.[0]= '\\' 
;;

let get_a_mask ms=
  let rec loop s acc= match s with
        [] -> (acc, s)
     |  (x::xs) when start_mask x -> (x::acc, xs)
     |  (x::xs) -> loop xs (x::acc)
  in
     loop ms []
;;

let rec prt_list s= match s with
     [] -> print_endline "Rest"; flush stdout
   | x::xs -> print_endline x; prt_list xs
;;

let get_masks ms=
   let rec loop s acc = match (get_a_mask s) with
         ([], []) -> []
      |  (xs, []) -> (the_mask xs)::acc
      |  (xs, rest) -> loop rest ((the_mask xs)::acc)
   in
      loop ms []
;;


let readMaskList fileName=
  let chan= open_in fileName and
      lines= ref [] in
  try
      while true do
         lines := (input_line chan)::lines.contents
      done;
       
      (get_masks lines.contents)
  with
     any -> get_masks lines.contents
;;

let mask_circle s=
  let ms= ref s in
     (fun () ->
        match ms.contents with
             [] when s= [] -> "no mask available"
          | [] -> ms := List.tl s; String.copy (List.hd s)
          | (x::xs) -> ms := xs; String.copy x
     )
;;
        